;; AGENDA (priority queue)
;;
;; OPERATIONS
;;
;;  (agenda.empty? a)        - boolean
;;  (agenda.time a)          - time of earliest item. Pre: not empty
;;  (agenda.remove-first a)  - remove and return first item. Pre: not empty
;;  (agenda.add a time item)
;;
;; Times are numbers.
;;
;; The agenda is STABLE, which means that items queued with the same time
;; are removed in the order that they were added.
;;
;; Note (1) users do not have to check if the agenda is empty every time.
;; The user can queue an item to do the check and use a non-local exit.

(declare (usual-integrations))

(define-macro (ucode-primitive . s) (apply make-primitive-procedure s))

(define-integrable skiplist-tag
  ((ucode-primitive string->symbol) "#[skiplist-agenda]"))

(named-structure/set-tag-description!
 skiplist-tag
 ((access make-define-structure-type system-global-environment)
  'vector
  "skiplist"
  '() '()
  ((access standard-unparser-method system-global-environment)
   'skiplist-agenda #f)))

;; slot   header  node
;;   0    tag     key
;;   1    level   datum
;;  2..   pointers
;;   n-1  random number seed (header only)

;; Level holds the index of the highest non-#F slot

(define (make-agenda)
  (vector skiplist-tag 1 #F #F #F #F #F #F #F #F #F #F #F 0))

(define (agenda.empty? agenda)
  (false? (vector-ref agenda 2)))

(define (agenda.time agenda)
  (vector-ref (vector-ref agenda 2) 0))

(define (agenda.remove-first agenda)
  (let ((first (vector-ref agenda 2)))
    (let ((first-level (fix:- (vector-length first) 1)))
      ;; Fixup header pointers
      (do ((i first-level (fix:- i 1)))
	  ((fix:= i 1))
	(vector-set! agenda i (vector-ref first i)))
      ;; Fixup list level
      (do ((i (vector-ref agenda 1) (fix:- i 1)))
	  ((vector-ref agenda i) (vector-set! agenda 1 i)))
      (vector-ref first 1))))

(define (agenda.add head time item)
  (let ((node (make-skiplist-node head time item)))
    (let ((list.level (vector-ref head 1))
	  (node.level (fix:- (vector-length node) 1)))
      (if (fix:> node.level list.level)
	  (vector-set! head 1 node.level))
      ;; search in taller nodes
      (let loop ((head* head) (i (vector-ref head 1)))
	(cond ((fix:= i node.level)
	       ;; search and insert
	       (let loop ((head* head*) (i i))
		 (cond ((fix:= i 1) 'done)
		       ((false? (vector-ref head* i))
			(vector-set! head* i node)
			(loop head* (fix:- i 1)))
		       ((<= (vector-ref (vector-ref head* i) 0) time)
			(loop (vector-ref head* i) i))
		       (else
			(vector-set! node i (vector-ref head* i))
			(vector-set! head* i node)
			(loop head* (fix:- i 1))))))
	      ((false? (vector-ref head* i))
	       (loop head* (fix:- i 1)))
	      ((<= (vector-ref (vector-ref head* i) 0) time)
	       (loop (vector-ref head* i) i))
	      (else
	       (loop head* (fix:- i 1))))))))


(define (make-skiplist-node head key datum)
  (let* ((rng-slot (fix:- (vector-length head) 1))
	 (rng (fix:and (fix:+ (vector-ref head rng-slot) 1)  #xffffff)))
    (define-integrable (test max-level k)
      (or (not (fix:= 0 (fix:and rng k)))
	  (fix:< rng-slot (fix:+ max-level 3))))
    (vector-set! head rng-slot rng)
    (cond
     ((test  1 #x3)     (vector key datum #F))
     ((test  2 #xf)     (vector key datum #F #F))
     ((test  3 #x3f)    (vector key datum #F #F #F))
     ((test  4 #xff)    (vector key datum #F #F #F #F))
     ((test  5 #x3ff)   (vector key datum #F #F #F #F #F))
     ((test  6 #xfff)   (vector key datum #F #F #F #F #F #F))
     ((test  7 #x3fff)  (vector key datum #F #F #F #F #F #F #F))
     ((test  8 #xffff)  (vector key datum #F #F #F #F #F #F #F #F))
     ((test  9 #x3ffff) (vector key datum #F #F #F #F #F #F #F #F #F))
     ((test 10 #xfffff) (vector key datum #F #F #F #F #F #F #F #F #F #F))
     (else	        (vector key datum #F #F #F #F #F #F #F #F #F #F #F)))))
