;;;; -*- Scheme -*-
;;
;;  $Id: comms3.scm,v 1.6 1997/09/25 15:52:05 adams Exp $
;;
;;  Copyright (c) 1997 Massachusetts Institute of Technology
;;  
;;  This material was developed by the Amorphous Computing project at the
;;  Massachusetts Institute of Technology Artificial Intelligence
;;  Laboratory.  Permission to copy this software, to redistribute it, and
;;  to use it for any purpose is granted, subject to the following
;;  restrictions and understandings.
;;  
;;  1. Any copy made of this software must include this copyright notice
;;  in full.
;;  
;;  2. Users of this software agree to make their best efforts (a) to
;;  return to the MIT Amorphous Computing project any improvements or
;;  extensions that they make, so that these may be included in future
;;  releases; and (b) to inform MIT of noteworthy uses of this software.
;;  
;;  3. All materials developed as a consequence of the use of this
;;  software shall duly acknowledge such use, in accordance with the usual
;;  standards of acknowledging credit in academic research.
;;  
;;  4. MIT has made no warrantee or representation that the operation of
;;  this software will be error-free, and MIT is under no obligation to
;;  provide any services, by way of maintenance, update, or otherwise.
;;  
;;  5. In conjunction with products arising from the use of this material,
;;  there shall be no use of the name of the Massachusetts Institute of
;;  Technology nor of any adaptation thereof in any advertising,
;;  promotional, or sales literature without prior written consent from
;;  MIT in each case.

(declare (usual-integrations))
(declare (integrate-external "sim"))

;; In this communications model it takes MESSAGE-DELAY units of time for
;; a message to be transmitted.  For a message to be detected its start
;; must be clear of any overlaps with other messages.  If several (two
;; or more) messages have any overlap then the first is delivered as
;; 'COLLISION and the subsequent messages are lost.

;; The comms slot is used on the receiver as follows. The scheme implicitly
;; assumes that all messages are exactly the same duration.  At
;; broadcast time the slot is filled with a unique message identifier.
;; At delivery time the slot is inspected.  If it has the same value
;; then there was no overlap and the message is delivered.  If it has
;; some other message identifier then we have detected a collision, and
;; 'COLLISION is delivered.  If the slot is #F then the message header
;; was lost so the message should be lost.  The slot is set to #F at
;; delivery time to signal that any currently active message must be
;; lost due to overlap with an earlier message.
;;
;; The slot also contains a `reference count' of all the current
;; transmissions.  This acts as a carrier-sense.

;; This communications model requires an agenda based scheduler
;; (e.g. simrun2)

(define message-delay 100)

(define (%broadcast cont fuel store message)

  (let* ((my-id (store.index store))
	 (sim   (store.simulation store))
	 (neighbours (vector-ref (simulation.neighbours sim) my-id))
	 (mark  (list my-id)))

    (define (delivery)
      ;; Use interrupt/1 as a reference/assignment cache for
      ;; PRIMITIVE-MESSAGE-EVENT
      (notify-delivery sim my-id message)
      (if (not (simulation.interrupt/1 sim))
	  (set-simulation.interrupt/1! sim
				       (make-cache 'PRIMITIVE-MESSAGE-EVENT)))
      (let ((event-cache (simulation.interrupt/1 sim)))
	(define (transmission-ended index)
	  (let ((store
		 (processor.store
		  (vector-ref (simulation.processors sim) index))))
	    (let ((event (global-ref store event-cache))
		  (comms-state (vector-ref store store:comms)))
	      (let ((mark* (car comms-state)))
		(set-car! comms-state #F)
		(set-cdr! comms-state (- (cdr comms-state) 1))
		(if (event? event)
		    (cond ((eq? mark* mark)
			   (event.signal! event message))
			  ((eq? mark* #F)
			   'message-lost)
			  (else		; overlap detected
			   (event.signal! event 'collision))))))))
	
	(for-each-vector-element neighbours transmission-ended)))

    (notify-broadcast sim my-id message)

    (for-each-vector-element neighbours
      (lambda (index)
	(let ((store
	       (processor.store
		(vector-ref (simulation.processors sim) index))))
	  (let ((comms-state (vector-ref store store:comms)))
	    (set-car! comms-state mark)
	    (set-cdr! comms-state (+ (cdr comms-state) 1))))))

    (simulation.enqueue+ sim message-delay delivery))

  (cont unspecific (fix:- fuel 1) store))


(define (processor.carrier-sense? processor)
  (let ((store (processor.store processor)))
    (let ((comms-state (vector-ref store store:comms)))
      (> (cdr comms-state) 0))))


(define (initial-comms-slot-value index)
  index ; ignored
  (cons #F 0))

(define (notify-broadcast sim my-id message)
  sim ; ignored
  (pp `(broadcast ,my-id ,@message)))

(define (notify-delivery sim my-id message)
  sim ; ignored
  (pp `(delivery ,my-id ,@message)))

