;;;; -*- Scheme -*-
;;
;;  Copyright (c) 1997 Massachusetts Institute of Technology
;;  
;;  This material was developed by the Amorphous Computing project at the
;;  Massachusetts Institute of Technology Artificial Intelligence
;;  Laboratory.  Permission to copy this software, to redistribute it, and
;;  to use it for any purpose is granted, subject to the following
;;  restrictions and understandings.
;;  
;;  1. Any copy made of this software must include this copyright notice
;;  in full.
;;  
;;  2. Users of this software agree to make their best efforts (a) to
;;  return to the MIT Amorphous Computing project any improvements or
;;  extensions that they make, so that these may be included in future
;;  releases; and (b) to inform MIT of noteworthy uses of this software.
;;  
;;  3. All materials developed as a consequence of the use of this
;;  software shall duly acknowledge such use, in accordance with the usual
;;  standards of acknowledging credit in academic research.
;;  
;;  4. MIT has made no warrantee or representation that the operation of
;;  this software will be error-free, and MIT is under no obligation to
;;  provide any services, by way of maintenance, update, or otherwise.
;;  
;;  5. In conjunction with products arising from the use of this material,
;;  there shall be no use of the name of the Massachusetts Institute of
;;  Technology nor of any adaptation thereof in any advertising,
;;  promotional, or sales literature without prior written consent from
;;  MIT in each case.

(declare (usual-integrations))

;; PNM utilities
;;
;; Utilities for simple operations on PBM, PGM and PPM files.
;;
;; (READ-PNM port-or-pathname)       read a PBM/PGM/PPM file
;; (PNM.WRITE pnm port-or-pathname)  write
;; (PNM.WIDTH pnm)                   size
;; (PNM.HEIGHT pnm)
;; (PNM.GET pnm x y)                 get pixel (native value)
;; (PNM.SET pnm x y value)           set pixel (native value)
;; (PNM.GET-RGB pnm x y)             get pixel as 24-bit RGB integer
;; (PNM.GET-BRIGHTNESS pnm x y)      get pixel as 0..1 flonum brightness
;; (MAKE-INITIALIZED-PBM width height procedure)
;; (MAKE-INITIALIZED-PGM width height procedure)
;; (MAKE-INITIALIZED-PPM width height procedure)
;;    make the appropriate PNM. PROCEDURE is called for every pixel
;;    and should return a native pixel value.

(define-structure
    (pnm
     (conc-name pnm.)
     (constructor %make-pnm (type width height depth comments data))
     (print-procedure
      (standard-unparser-method 'PNM
	(lambda (pnm port)
	  (write-char #\space port)
	  (write (pnm.width pnm) port)
	  (write-char #\x port)
	  (write (pnm.height pnm) port)
	  (write-char #\space port)
	  (write-string (pnm-type.name (pnm.type pnm)) port)))))

  (type    #f read-only true)
  (width   0  read-only true)
  (height  0  read-only true)
  (depth   0  read-only true)
  (comments '())
  (data    #f read-only true))

(define-structure
    (pnm-type
     (conc-name pnm-type.)
     (print-procedure
      (standard-unparser-method 'PNM-TYPE
	(lambda (type port)
	  (write-char #\space port)
	  (write-string (pnm-type.name type) port)))))
  (name  #f  read-only true)
  (type  #F  read-only true)            ; 'PBM or 'PGM or 'PPM
  (get   #f  read-only true)		; (lambda (pnm x y) ...)  -> pixel
  (set   #f  read-only true)		; (lambda (pnm x y v) ...)
  (get-brightness #f read-only true)	; (lambda (pnm x y) ...) -> flonum
  (get-rgb #f  read-only true)		; (lambda (pnm x y) ...) -> color
  (write   #f  read-only true)		; (lambda (pnm port) ...)
  (operations '() read-only true))	; alist


(define (pbm? object)
  (and (pnm? object)
       (eq? (pnm-type.type (pnm.type object)) 'PBM)))

(define (pgm? object)
  (and (pnm? object)
       (eq? (pnm-type.type (pnm.type object)) 'PGM)))

(define (ppm? object)
  (and (pnm? object)
       (eq? (pnm-type.type (pnm.type object)) 'PPM)))

(define (pnm.get pnm x y)
  (guarantee-pnm pnm 'PNM.GET)
  ((pnm-type.get (pnm.type pnm)) pnm x y))

(define (pnm.set pnm x y pixel)
  (guarantee-pnm pnm 'PNM.SET)
  ((pnm-type.set (pnm.type pnm)) pnm x y pixel))

(define (pnm.get-brightness pnm x y)
  ;; returns a flonum between 0 and 1. inclusive
  (guarantee-pnm pnm 'PNM.GET-BRIGHTNESS)
  ((pnm-type.get-brightness (pnm.type pnm)) pnm x y))

(define (pnm.get-rgb pnm x y)
  ;; returns a color #xrrggbb
  (guarantee-pnm pnm 'PNM.GET-RGB)
  ((pnm-type.get-rgb (pnm.type pnm)) pnm x y))


(define (pnm.write pnm port)
  (define (write port)
    ((pnm-type.write (pnm.type pnm)) pnm port))
  (guarantee-pnm pnm 'PNM.WRITE)
  (cond ((port? port) (write port))
	((or (string? port) (pathname? port))
	 (call-with-binary-output-file port write))
	(else
	 (error:wrong-type-argument port "port or pathname" 'PNM.WRITE))))

(define (pnm.operation pnm operation . arguments)
  (guarantee-pnm pnm 'PNM.OPERATION)
  (cond ((assq operation (pnm.operations pnm))
	 => (lambda (pair)
	      (apply (cdr pair) arguments)))
	(else
	 (error "Not a valid PNM operation" operation pnm))))

(define (pnm.add-comment pnm string)
  (guarantee-pnm pnm 'PNM.ADD-COMMENT)
  (guarantee-string string 'PNM.ADD-COMMENT)
  (pnm.add-comments pnm (list string)))

(define (pnm.add-comments pnm strings)
  (guarantee-pnm pnm 'PNM.ADD-COMMENTS)
  (if (not (and (list? strings)
		(for-all? strings string?)))
      (error:wrong-type-argument strings "list of strings" 'PNM.ADD-COMMENTS))
  (set-pnm.comments pnm (append (pnm.comments pnm) strings)))



(define (make-pbm width height)
  (%make-pnm binary-pbm-type width height 1
	     (list (string-append " Created: "
				  (decoded-time->string (get-decoded-time))))
	     (make-string (quotient (+ 7 (* width height)) 8) #\nul)))

(define (make-pgm width height)
  (%make-pnm binary-pgm-type width height 255
	     (list (string-append " Created: "
				  (decoded-time->string (get-decoded-time))))
	     (make-string (* width height) #\nul)))

(define (make-ppm width height)
  (%make-pnm binary-ppm-type width height 255
	     (list (string-append " Created: "
				  (decoded-time->string (get-decoded-time))))
	     (make-string (* 3 width height) #\nul)))


(define (initialize-pnm-from-function pnm function)
  ;; returns PNM
  (let ((width  (pnm.width pnm))
	(height (pnm.height pnm))
	(setter (pnm-type.set (pnm.type pnm))))
    (do ((j 0 (fix:+ j 1)))
	((fix:= j height))
      (do ((i 0 (fix:+ i 1)))
	  ((fix:= i width))
	(setter pnm i j (function i j)))))
  pnm)
      
(define (make-initialized-pbm width height function)
  (initialize-pnm-from-function (make-pbm width height) function))

(define (make-initialized-pgm width height function)
  (initialize-pnm-from-function (make-pgm width height) function))

(define (make-initialized-ppm width height function)
  (initialize-pnm-from-function (make-ppm width height) function))



(define (read-pnm port)

  (define comments '())

  (define (read-binary-ppm)
    (read-header
     (lambda (width height depth comments)
       (if (not (= depth 255))
	   (error "Sorry, only know about 256-level PPMs" depth port))
       (let ((data (read-string (char-set) port)))
	 (let ((expected-length (* 3 width height))
	       (actual-length (string-length data)))
	   (if (= expected-length actual-length)
	       (%make-pnm binary-ppm-type width height depth comments data)
	       (error "Bad data - expected" expected-length
		      (error-irritant/noise "bytes, got")
		      actual-length port)))))))

  (define (read-binary-pgm)
    (read-header
     (lambda (width height depth comments)
       (if (not (= depth 255))
	   (error "Sorry, only know about 256-level PGMs" depth port))
       (let ((data (read-string (char-set) port)))
	 (let ((expected-length (* width height))
	       (actual-length (string-length data)))
	   (if (= expected-length actual-length)
	       (%make-pnm binary-pgm-type width height depth comments data)
	       (error "Bad data - expected" expected-length
		      (error-irritant/noise "bytes, got")
		      actual-length port)))))))

  (define (read-binary-pbm)
    (read-pbm-header
     (lambda (width height comments)
       (let ((data (read-string (char-set) port)))
	 (let ((expected-length (quotient (+ 7 (* width height)) 8))
	       (actual-length (string-length data)))
	   (if (= expected-length actual-length)
	       (%make-pnm binary-pbm-type width height 1 comments data)
	       (error "Bad data - expected" expected-length
			     (error-irritant/noise "bytes, got")
			     actual-length port)))))))

  (define (read-header receiver)
    (set! comments '())
    (read-number
     (lambda (width)
       (read-number
	(lambda (height)
	  (read-number
	   (lambda (max-value)
	     (receiver width height max-value (reverse comments)))))))))

  (define (read-pbm-header receiver)
    (set! comments '())
    (read-number
     (lambda (width)
       (read-number
	(lambda (height)
	  (receiver width height (reverse comments)))))))

  (define (read-number receiver)
    (let loop ()
      (let ((ch (read-char port)))
	(case ch
	  ((#\space #\newline #\return) (loop))
	  ((#\#)     (read-comment loop))
	  ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)
	   (let digit-loop ((n (char->digit ch)))
	     (let ((ch (read-char port)))
	       (case ch
		 ((#\space #\newline #\return) (receiver n))
		 ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9)
		  (digit-loop (+ (* n 10) (char->digit ch))))
		 (else
		  (error "Wierd character after number" ch n port))))))
	  (else
	   (error "Illegal character" ch port))))))

  (define (read-comment continue)
    (let loop ((chars '()))
      (let ((ch (read-char port)))
	(case ch
	  ((#\return #\linefeed)
	   (set! comments (cons (list->string (reverse chars)) comments))
	   (continue))
	  (else
	   (loop (cons ch chars)))))))

  (cond ((port? port)
	 (let ((ch (read-char port)))
	   (if (eqv? ch #\P)
	       (let ((type (read-char port)))
		 (case type
		   ((#\1) (error "Sorry, ASCII PBMs not supported" port))
		   ((#\2) (error "Sorry, ASCII PGMs not supported" port))
		   ((#\3) (error "Sorry, ASCII PPMs not supported" port))
		   ((#\4) (read-binary-pbm))
		   ((#\5) (read-binary-pgm))
		   ((#\6) (read-binary-ppm))
		   (else (error "Unknown PNM type" (string #\P type) port))))
	       (error "Not a PNM file:" port))))
	((or (string? port)
	     (pathname? port))
	 (call-with-binary-input-file port read-pnm))
	(else
	 (error:wrong-type-argument port "port or pathname" 'READ-PNM))))

(define (write-binary-pnm pnm type-code write-depth? port)
  (write-string type-code port)
  (write-char #\newline port)
  (for-each
      (lambda (s)
	(write-char #\# port)
	(write-string s port)
	(write-char #\newline port))
    (reverse (pnm.comments pnm)))
  (write (pnm.width pnm) port)
  (write-char #\space port)
  (write (pnm.height pnm) port)
  (write-char #\newline port)
  (if write-depth?
      (begin
	(write (pnm.depth pnm) port)
	(write-char #\newline port)))
  (write-string (pnm.data pnm) port))


(define binary-pbm-type)
(define binary-pgm-type)
(define binary-ppm-type)

(define (initialize!)

  (define (make-binary-pbm-type)
    (define-integrable (loc pnm x y receiver)
      (let ((bit-number (fix:+ x (fix:* y (pnm.width pnm)))))
	(receiver (fix:quotient bit-number 8)
		  (fix:lsh 1 (fix:and bit-number #x7)))))

    (define-integrable (get-choice pnm x y zero one)
      (loc pnm x y
	   (lambda (index mask)
	     (if (= 0 (fix:and (vector-8b-ref (pnm.data pnm) index) mask))
		 zero
		 one))))

    (define (get pnm x y) (get-choice pnm x y 0 1))

    (define (set pnm x y level)
      (loc pnm x y
	   (lambda (index mask)
	     (if (or (eq? level #F) (eq? level 0))
		 (vector-8b-set! (pnm.data pnm) index
				 (fix:andc (vector-8b-ref (pnm.data pnm) index)
					   mask))
		 (vector-8b-set! (pnm.data pnm) index
				 (fix:or (vector-8b-ref (pnm.data pnm) index)
					 mask))))))

    (define (get-brightness pnm x y) (get-choice pnm x y 0.0 1.0))

    (define (get-rgb pnm x y) (get-choice pnm x y #x000000 #xffffff))

    (define (write pnm port)
      (write-binary-pnm pnm "P4" #f port))

    (make-pnm-type "binary PBM" 'PBM get set get-brightness get-rgb write '()))


  (define (make-binary-pgm-type)
    (define-integrable (loc pnm x y)
      (fix:+ x (fix:* y (pnm.width pnm))))

    (define (get pnm x y)
      (vector-8b-ref (pnm.data pnm) (loc pnm x y)))

    (define (set pnm x y level)
      (vector-8b-set! (pnm.data pnm) (loc pnm x y) level))
   
    (define (get-brightness pnm x y)
      (* (/ 1 255.) (get pnm x y)))

    (define (get-rgb pnm x y)
      (let ((level (get pnm x y)))
	(+ (fix:lsh level 16) (fix:lsh level 8) level)))

    (define (write pnm port)
      (write-binary-pnm pnm "P5" #t port))

    (make-pnm-type "binary PGM" 'PGM get set get-brightness get-rgb write '()))


  (define (make-binary-ppm-type)
    (define-integrable (loc pnm x y)
      (fix:* 3 (fix:+ x (fix:* y (pnm.width pnm)))))

    (define (get pnm x y)
      (let ((p (loc pnm x y)))
	(fix:+ (fix:lsh (vector-8b-ref (pnm.data pnm) p) 16)
	       (fix:+ (fix:lsh (vector-8b-ref (pnm.data pnm) (fix:+ p 1)) 8)
		      (vector-8b-ref (pnm.data pnm) (fix:+ p 2))))))

    (define (set pnm x y rgb)
      (let ((p (loc pnm x y)))
	(vector-8b-set! (pnm.data pnm) (fix:+ p 0) (fix:lsh rgb -16))
	(vector-8b-set! (pnm.data pnm) (fix:+ p 1) (fix:and (fix:lsh rgb -8) 255))
	(vector-8b-set! (pnm.data pnm) (fix:+ p 2) (fix:and rgb 255))))

    (define (get-brightness pnm x y)
      (let ((p (loc pnm x y)))
	(* (/ 1. (* 3 255))
	   (fix:+ (vector-8b-ref (pnm.data pnm) p)
		  (fix:+ (vector-8b-ref (pnm.data pnm) (fix:+ p 1))
			 (vector-8b-ref (pnm.data pnm) (fix:+ p 2)))))))

    (define (write pnm port)
      (write-binary-pnm pnm "P6" #t port))

    (make-pnm-type "binary PPM" 'PPM get set get-brightness get write '()))

  (set! binary-pbm-type (make-binary-pbm-type))
  (set! binary-pgm-type (make-binary-pgm-type))
  (set! binary-ppm-type (make-binary-ppm-type))
  unspecific)


(define-integrable (guarantee-pnm object procedure)
  (if (not (pnm? object))
      (error:wrong-type-argument object "portable `any'-map (PNM)" procedure)))

(initialize!)
